@@ -11,12 +11,14 @@
 # ANY KIND, either express or implied. See the License for the specific
 # language governing permissions and limitations under the License.
 
+import numbers
 from functools import reduce, singledispatch
-from typing import Any, FrozenSet, List, Tuple, Union
+from typing import Any, FrozenSet, List, Optional, Tuple, Union
 
 import pennylane as qml
-from braket.circuits import Gate, ResultType, gates, noises, observables
+from braket.circuits import FreeParameter, Gate, ResultType, gates, noises, observables
 from braket.circuits.result_types import (
+    AdjointGradient,
     DensityMatrix,
     Expectation,
     Probability,
@@ -30,9 +32,18 @@ from pennylane import numpy as np
 from pennylane.measurements import ObservableReturnTypes
 from pennylane.operation import Observable, Operation
 
-from braket.pennylane_plugin.ops import PSWAP, CPhaseShift00, CPhaseShift01, CPhaseShift10
+from braket.pennylane_plugin.ops import (
+    MS,
+    PSWAP,
+    CPhaseShift00,
+    CPhaseShift01,
+    CPhaseShift10,
+    GPi,
+    GPi2,
+)
 
 _BRAKET_TO_PENNYLANE_OPERATIONS = {
+    "i": "Identity",
     "x": "PauliX",
     "y": "PauliY",
     "z": "PauliZ",
@@ -72,6 +83,9 @@ _BRAKET_TO_PENNYLANE_OPERATIONS = {
     "yy": "IsingYY",
     "zz": "IsingZZ",
     "ecr": "ECR",
+    "gpi": "GPi",
+    "gpi2": "GPi2",
+    "ms": "MS",
 }
 
 
@@ -89,23 +103,57 @@ def supported_operations(device: Device) -> FrozenSet[str]:
     except AttributeError:
         raise AttributeError("Device needs to have properties defined.")
     supported_ops = frozenset(op.lower() for op in properties.supportedOperations)
+    supported_pragmas = frozenset(op.lower() for op in properties.supportedPragmas)
     return frozenset(
         _BRAKET_TO_PENNYLANE_OPERATIONS[op]
         for op in _BRAKET_TO_PENNYLANE_OPERATIONS
-        if op.lower() in supported_ops
+        if op.lower() in supported_ops or f"braket_noise_{op.lower()}" in supported_pragmas
     )
 
 
-def translate_operation(operation: Operation, *args, **kwargs) -> Gate:
+def translate_operation(
+    operation: Operation,
+    use_unique_params: bool = False,
+    param_names: Optional[List[str]] = None,
+    *args,
+    **kwargs,
+) -> Gate:
     """Translates a PennyLane ``Operation`` into the corresponding Braket ``Gate``.
 
     Args:
         operation (Operation): The PennyLane ``Operation`` to translate
+        use_unique_params (bool): If true, numeric parameters in the resulting operation will be
+        replaced with FreeParameter objects (with names corresponding to param_names). Non-numeric
+        parameters will be skipped.
+        param_names (Optional[List[str]]): A list of parameter names
+            to be supplied to the new operation.
 
     Returns:
         Gate: The Braket gate corresponding to the given operation
     """
-    parameters = [p.numpy() if isinstance(p, qml.numpy.tensor) else p for p in operation.parameters]
+    if use_unique_params:
+        param_names = param_names or []
+        parameters = []
+        name_index = 0
+        for param in operation.parameters:
+            # PennyLane passes any non-keyword argument in the operation.parameters list.
+            # In some cases, like the unitary gate or qml.QubitChannel (Kraus noise), these
+            # parameter can be matrices. Braket only supports parameterization of numeric parameters
+            # (so far, these are all angle parameters), so non-numeric parameters are handled
+            # separately.
+            if isinstance(param, numbers.Number):
+                new_param = FreeParameter(param_names[name_index])
+                name_index += 1
+            elif isinstance(param, qml.numpy.tensor):
+                new_param = param.numpy()
+            else:
+                new_param = param
+            parameters.append(new_param)
+    else:
+        parameters = [
+            p.numpy() if isinstance(p, qml.numpy.tensor) else p for p in operation.parameters
+        ]
+
     return _translate_operation(operation, parameters)
 
 
@@ -116,6 +164,11 @@ def _translate_operation(operation: Operation, _parameters) -> Gate:
     )
 
 
+@_translate_operation.register
+def _(_: qml.Identity, _parameters):
+    return gates.I()
+
+
 @_translate_operation.register
 def _(_: qml.Hadamard, _parameters):
     return gates.H()
@@ -318,6 +371,40 @@ def _(zz: qml.IsingZZ, parameters):
     return gates.ZZ(-phi) if zz.inverse else gates.ZZ(phi)
 
 
+@_translate_operation.register
+def _(_gpi: GPi, parameters):
+    phi = parameters[0]
+    return gates.GPi(phi)
+
+
+@_translate_operation.register
+def _(gpi2: GPi2, parameters):
+    phi = parameters[0]
+    return gates.GPi2(phi + np.pi) if gpi2.inverse else gates.GPi2(phi)
+
+
+@_translate_operation.register
+def _(ms: MS, parameters):
+    phi_0, phi_1 = parameters[:2]
+    return gates.MS(phi_0 + np.pi, phi_1) if ms.inverse else gates.MS(phi_0, phi_1)
+
+
+def get_adjoint_gradient_result_type(
+    observable: Observable,
+    targets: Union[List[int], List[List[int]]],
+    supported_result_types: FrozenSet[str],
+    parameters: List[str],
+):
+    if "AdjointGradient" not in supported_result_types:
+        raise NotImplementedError("Unsupported return type: AdjointGradient")
+    braket_observable = _translate_observable(observable)
+
+    braket_observable = (
+        braket_observable.item() if hasattr(braket_observable, "item") else braket_observable
+    )
+    return AdjointGradient(observable=braket_observable, target=targets, parameters=parameters)
+
+
 def translate_result_type(
     observable: Observable, targets: List[int], supported_result_types: FrozenSet[str]
 ) -> Union[ResultType, Tuple[ResultType, ...]]:
@@ -366,7 +453,21 @@ def translate_result_type(
 
 @singledispatch
 def _translate_observable(observable):
-    raise TypeError(f"Unsupported observable: {observable}")
+    raise TypeError(f"Unsupported observable: {type(observable)}")
+
+
+@_translate_observable.register
+def _(H: qml.Hamiltonian):
+    # terms is structured like [C, O] where C is a tuple of all the coefficients, and O is
+    # a tuple of all the corresponding observable terms (X, Y, Z, H, etc or a tensor product
+    # of them)
+    coefficents, pl_observables = H.terms()
+    braket_observables = list(map(lambda obs: _translate_observable(obs), pl_observables))
+    braket_hamiltonian = sum(
+        (coef * obs for coef, obs in zip(coefficents[1:], braket_observables[1:])),
+        coefficents[0] * braket_observables[0],
+    )
+    return braket_hamiltonian
 
 
 @_translate_observable.register
@@ -437,6 +538,20 @@ def translate_result(
     Note:
         Hamiltonian results will be summed over all terms.
     """
+
+    # if braket result contains adjoint gradient, just return it since it should be the only
+    # result type if it's there at all.
+    ag_results = [
+        result for result in braket_result.result_types if result.type.type == "adjoint_gradient"
+    ]
+    if ag_results:
+        ag_result = ag_results[0]
+        key_indices = [int(param_name.split("p_")[1]) for param_name in ag_result.value["gradient"]]
+        return [ag_result.value["expectation"]], [
+            # we need to sort the keys by index since braket can return them in the wrong order
+            ag_result.value["gradient"][f"p_{i}"]
+            for i in sorted(key_indices)
+        ]
     translated = translate_result_type(observable, targets, supported_result_types)
     if isinstance(observable, qml.Hamiltonian):
         coeffs, _ = observable.terms()
